
#ifndef SIER_H

#define SIER_H

/*===============================<o>=====================================

  This software was written by Ian L. Kaplan, Chief Fat Bear, Bear Products
  International.  Use of this software, for any purpose, is granted on two
  conditions: 

   1. This copyright notice must be included with the software
      or any software derived from it.

   2. The risk of using this software is accepted by the user.  No
      warranty as to its usefulness or functionality is provided
      or implied.  The author and Bear Products International provides
      no support.

   May 1996

 ===============================<o>=====================================*/


#include <memory.h>    // needed for memcpy
#include <malloc.h>    // needed for malloc
#include "pagepool.h"  // needed for the pagepool memory allocator


class three_D_point {
 private:
  typedef enum { Xcoord = 0, Ycoord = 1, Zcoord = 2 } coords;
  int vect[ 3 ];
 public:
  void set_point( int X, int Y, int Z ) { vect[ Xcoord ] = X;
					  vect[ Ycoord ] = Y;
					  vect[ Zcoord ] = Z; }
  int getX(void) { return vect[ Xcoord ]; }
  int getY(void) { return vect[ Ycoord ]; }
  int getZ(void) { return vect[ Zcoord ]; }
  void pointcopy( int *&destaddr ) { 
    memcpy(destaddr, vect, sizeof(vect));  
    destaddr += sizeof(vect )/sizeof(int); 
  } 
};



/* A four sided pyramid is defined by five points: The four points making 
   up the base (A, B, C, D) and Apex
 */
class pyramid {
 protected:
  typedef enum { A, B, C, D, Apex } pyramid_enum;
  typedef enum { NumPoints = 5 } pyramid_const;
  three_D_point pyramid_points[ NumPoints ];
 public:
   void pyr_get_faces(int *&list );
   pyramid( three_D_point apex );
};



class display_pyramid : public pyramid {
private:
  void print_indent(int indent );
public:
  display_pyramid( three_D_point apex ) : pyramid( apex ) {};
  void debug_print(int indent );
}; // display_pyramid


/* The Sierpinski pyramid is recursively built.  For a Sierpinski pyramid
   of power N, an N-1 pyramid is supported by four N-1 children.  The
   building block for the Sierpinski pyramid is a four sided pyramid.
*/



//
// class Sierpinski_types
//
// data types for Sierpinski pyramid
//
class Sierpinski_types {
 public:
  typedef enum { height = 2,       // height of pyramid
                 NumChildren = 4,  // Number of pyramid children
                 MaxPower = 6      // maximum "power" for Sierpinski pyramid
	       };

  typedef enum { pyramid_object, sierpinski_object } sierpinski_kind;

 //
 // Protected enumerations and types
 //
 protected:
  typedef enum {points_per_face = 4,
	        faces_per_pyramid = 5
               };

  typedef int point[ 3 ];
  typedef point face[points_per_face];


  typedef display_pyramid *pyramid_ptr;

  typedef struct sier_struct {
    sierpinski_kind obj_kind;  // tag for union
    union { // "naked" union 
      pyramid_ptr pyr;
      sier_struct *sier;
    };
    sier_struct *children[ NumChildren ];
  } sierpinski_struct;

  typedef sierpinski_struct *sierpinski_ptr;
};  // Sierpinski_types



//
// class Sierpinski
//
class Sierpinski : public Sierpinski_types {

 // Private class variables
 private:
   // The memory allocator object is declared static, which allows
   // all objects of the class to share the same memory allocator object.
   static page_pool mem;


 // Private class functions
 private:
   three_D_point apex_calc( three_D_point apex, 
		            int power, 
			    int corner );

   int power_calc( int power );
   sierpinski_ptr build(three_D_point apex, int power);
   void get_faces( sierpinski_ptr root, int *&list );
   int calc_num_faces(void);

 // Protected class variables
 protected:
  // root is the apex of the Sierpinski pyramid.  The pyramid is
  // built by the constructor.
   sierpinski_ptr root;

   int *facelist;
   int num_faces;
   int pyr_power;  // power of two for the Sierpinski pyramid

 // Public class functions
 public:
    Sierpinski( int power );
    ~Sierpinski(void) { free( facelist ); }
    int get_size() { return power_calc( pyr_power + 1 ); }
    int *get_facelist(void) { return facelist; };
    int get_num_faces(void) { return num_faces; };
}; // Sierpinski



//
// class debug_sierpinski
//
// This class inherits the Sierpinski class and adds functions
// to print the Sierpinski data structure and to print the
// "facelist" array.
//
class debug_sierpinski : public Sierpinski
{

 private:
   void sier_debug_print(sierpinski_ptr root, int level );
   int point_neq(int *a, int *b) { return (a[0] != b[0] || 
					   a[1] != b[1] || 
					   a[2] != b[2]);
				 }

 public:
  void debug_print(void) { sier_debug_print( root, 0 ); }
  void print_facelist(void);
  debug_sierpinski( int power ) : Sierpinski(power) {};
}; // display_sierpinski


#endif

